#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#include "wllib.h"

#define SAFE_BORDER (4)
#define SIZE_MIN (16)

extern char *icon_xpm[];
extern void drawgfx(void);

long xcur,ycur,xmin,xmax,ymin,ymax;
int winx,winy,iniline,minimax=1;
float coefx,coefy;

Display *dp;
Screen *scr;
int scrnum;
Colormap cmap;
unsigned long fgpen,bgpen;
XrmDatabase rdb;
Window win,rootw;
GC gc;
Cursor curs;
Atom atdel;

XrmOptionDescRec opts[]={
{"-bg",          ".background",  XrmoptionSepArg,NULL}, /*  0! */
{"-background",  ".background",  XrmoptionSepArg,NULL}, /*  1! */
{"-bd",          ".borderColor", XrmoptionSepArg,NULL}, /*  2! */
{"-bordercolor", ".borderColor", XrmoptionSepArg,NULL}, /*  3! */
{"-bw",          ".borderWidth", XrmoptionSepArg,NULL}, /*  4! */
{"-borderwidth", ".borderWidth", XrmoptionSepArg,NULL}, /*  5! */
{"-display",     ".display",     XrmoptionSepArg,NULL}, /*  6! */
{"-fg",          ".foreground",  XrmoptionSepArg,NULL}, /*  7! */
{"-foreground",  ".foreground",  XrmoptionSepArg,NULL}, /*  8! */
{"-geometry",    ".geometry",    XrmoptionSepArg,NULL}, /*  9! */
{"-iconic",      ".iconStartup", XrmoptionNoArg, "on"}, /* 10! */
{"-name",        ".name",        XrmoptionSepArg,NULL}, /* 11! */
{"-rv",          ".reverseVideo",XrmoptionNoArg, "on"}, /* 12! */
{"-reverse",     ".reverseVideo",XrmoptionNoArg, "on"}, /* 13! */
{"+rv",          ".reverseVideo",XrmoptionNoArg, "on"}, /* 14! */
{"+reverse",     ".reverseVideo",XrmoptionNoArg, "on"}, /* 15! */
{"-title",       ".title",       XrmoptionSepArg,NULL}, /* 16! */
{"-xrm",         NULL,           XrmoptionResArg,NULL}, /* 17! */
};
#define opts_num (sizeof(opts)/sizeof(*opts))

char *pname;
size_t pnamelen;
char fn[128]="/usr/lib/X11/app-defaults/",resnam[sizeof(fn)],rescls[sizeof(fn)];
#define APPDEFNL (26)
size_t classlen;

void wl_line(long x,long y)
{
	if (iniline) {
		iniline=0;
		if (minimax) { xmin=xmax=x; ymin=ymax=y; }
		else {
			xcur=rint(coefx*(x-xmin))+SAFE_BORDER;
			ycur=winy-SAFE_BORDER-rint(coefy*(y-ymin));
			}
		}
	else {
		if (minimax) {
			if (x<xmin) xmin=x; if (x>xmax) xmax=x;
			if (y<ymin) ymin=y; if (y>ymax) ymax=y;
			}
		else {
int xcur2,ycur2;
			XDrawLine(dp,win,gc,xcur,ycur,
					xcur2=rint(coefx*(x-xmin))+SAFE_BORDER,ycur2=winy-SAFE_BORDER-rint(coefy*(y-ymin)));
			xcur=xcur2; ycur=ycur2;
			}
		}
}

int stricmp(char *s1,char *s2)
{
int i;
char c1,c2;
	while (*s1&&*s2) if ((i=(c1=toupper(*s1++))>(c2=toupper(*s2++))-(c1<c2))) return(i);
	return(0);
}

static char *res_string(int resn,const char *resc)
{
char *s;
XrmValue val;
	strncpy(&resnam[pnamelen],opts[resn].specifier,sizeof(resnam)-pnamelen);
	strncpy(&rescls[classlen],resc                ,sizeof(rescls)-classlen);
	if (XrmGetResource(rdb,resnam,rescls,&s,&val)) {
		if (!(s=malloc(val.size))) iofail();
		return(strncpy(s,val.addr,val.size));
		}
	return(NULL);
}

static int res_bool(char *s)
{
static char *yestr[4]={"1","on","yes","true"};
#define YESTRL (sizeof(yestr)/sizeof(*yestr))
int i;
	if (s) for (i=0;i<YESTRL;i++)
		if (!stricmp(s,yestr[i])) { free(s); return(1); }
	free(s);
	return(0);
}

static int res_int(char *s,int def)
{
char *ers;
	if (s) {
		def=strtol(s,&ers,0);
		if (ers&&*ers) { fprintf(stderr,"strtol() error @\"%s\"\n",ers); iofail(); }
		free(s);
		}
	return(def);
}

typedef enum { color_def_White,color_def_Black } color_def;
static unsigned long get_color(char *desc,color_def def)
{
XColor xcol_scr,xcol_ex;
	if (!desc) return(def==color_def_White?WhitePixelOfScreen(scr):BlackPixelOfScreen(scr));
	else if (!XAllocNamedColor(dp,cmap=DefaultColormapOfScreen(scr),desc,&xcol_scr,&xcol_ex))
			{ fprintf(stderr,"XAllocNamedColor(\"%s\")\n",desc); iofail(); }
	return(xcol_scr.pixel);
}

void wl_done(void)
{
XEvent ev;
	iniline=1; drawgfx();
	minimax=0;
	if (xmax==xmin) xmax++;
	if (ymax==ymin) ymax++;
	XUndefineCursor(dp,win);
	XFreeCursor(dp,curs);
	for (;;) {
		XNextEvent(dp,&ev);
		if (ev.xany.window==win) switch(ev.type) {
			case ConfigureNotify:
				winx=ev.xconfigure.width; winy=ev.xconfigure.height;
				break;
			case ClientMessage:
				if (ev.xclient.format==32) if (ev.xclient.data.l[0]==atdel)
			case ButtonPress:
			case KeyPress:
					exit(EXIT_SUCCESS);
				break;
			case Expose:
				XClearArea(dp,win,ev.xexpose.x,ev.xexpose.y,ev.xexpose.width,ev.xexpose.height,False);
				if (ev.xexpose.count) break;
				if (!iniline) {
					coefx=(float)(winx-1-2*SAFE_BORDER)/(xmax-xmin);
					coefy=(float)(winy-1-2*SAFE_BORDER)/(ymax-ymin);
					iniline=1; drawgfx();
					}
			}
		}
}

void wl_init(int *argcp,char **argv,const char *class)
{
char *s,*disp;
size_t ts;
unsigned long bdpen;
unsigned bw;
XSetWindowAttributes wina;
XSizeHints *sizh;
XWMHints *wmh;
XClassHint *clsh;
int xs,ys,i;
struct {
	Pixmap data,mask;
	XpmAttributes attr;
	Window win;
	XIconSize *siz;
	} icon;

	if ((pname=rindex(*argv,'/'))) pname++;
	else pname=*argv;
	pnamelen=strlen(pname);
	XrmInitialize();
	strncpy(&fn[APPDEFNL],class,sizeof(fn)-APPDEFNL);
	rdb=XrmGetFileDatabase(fn);
	if (!(s=getenv("XUSERFILESEARCHPATH"))) s=getenv("XAPPLRESDIR");
	if (s) if (!(ts=strlen(strncpy(fn,s,sizeof(fn))))) {
		if (fn[ts-1]!='/') fn[ts++]='/';
		strncpy(&fn[ts],class,sizeof(fn)-ts);
		XrmCombineFileDatabase(fn,&rdb,True);
		}
	XrmParseCommand(&rdb,opts,opts_num,class,argcp,argv);
	strncpy(resnam,pname,sizeof(resnam));
	strncpy(rescls,class,sizeof(rescls));
	classlen=strlen(class);
	if ((s=res_string(11,".Name")))
		pnamelen=strlen(pname=s);
	if (!(dp=XOpenDisplay(disp=res_string(6,".Display"))))
		{ fprintf(stderr,"Display=\"%s\"\n",XDisplayName(disp)); iofail(); }
	if (!(s=getenv("HOME"))) s="~";
	ts=strlen(strncpy(fn,s,sizeof(fn)));
	strncpy(&fn[ts],"/.Xdefaults-",sizeof(fn)-ts); ts+=12;
	if (!(s=getenv("XENVIRONMENT"))) {
		if (gethostname(&fn[ts],sizeof(fn)-ts)) iofail();
		s=fn;
		}
	XrmCombineFileDatabase(s,&rdb,False);
	if ((s=XScreenResourceString(scr=ScreenOfDisplay(dp,scrnum=DefaultScreen(dp)))))
		XrmCombineDatabase(XrmGetStringDatabase(s),&rdb,False);
	if ((s=XResourceManagerString(dp)))
		XrmCombineDatabase(XrmGetStringDatabase(s),&rdb,False);
	else {
		fn[ts-1]='\0';
		XrmCombineFileDatabase(fn,&rdb,False);
		}
	bw=res_int(res_string(4,".BorderWidth"),0);
	bgpen=get_color(res_string(0,".Background" ),color_def_White);
	fgpen=get_color(res_string(7,".Foreground" ),color_def_Black);
	if (res_bool(res_string(12,".ReverseVideo")))
		{ bdpen=bgpen; bgpen=fgpen; fgpen=bdpen; }
	bdpen=get_color(res_string(2,".BorderColor"),color_def_Black);
	if (!(sizh=XAllocSizeHints())) iofail();
	xs=WidthOfScreen(scr)/2; ys=HeightOfScreen(scr)/2;
	sprintf(fn,"%dx%d+%d+%d",xs,ys,xs/2,ys/2);
	sizh->flags=(i=XWMGeometry(dp,scrnum,res_string(9,".Geometry"),fn,bw,sizh,
					&sizh->x,&sizh->y,&winx,&winy,&sizh->win_gravity)
			&(XValue|YValue)?USPosition:PPosition)|(i&(WidthValue|HeightValue)?USSize:PSize)
			|PMinSize|PWinGravity;
	sizh->min_height=sizh->min_width=2*SAFE_BORDER+SIZE_MIN;
	if (!(win=XCreateSimpleWindow(dp,rootw=RootWindowOfScreen(scr),
			sizh->x,sizh->y,sizh->width=winx,sizh->height=winy,bw,bdpen,bgpen))) iofail();
	icon.attr.valuemask=0;
	if ((i=XpmCreatePixmapFromData(dp,win,icon_xpm,&icon.data,&icon.mask,&icon.attr)))
		{ fprintf(stderr,"XpmCreatePixmapFromData() error #%d\n",i); iofail(); }
	if (!(icon.siz=XAllocIconSize())) iofail();
	icon.siz->max_width =icon.siz->min_width =icon.attr.width;
	icon.siz->max_height=icon.siz->min_height=icon.attr.height;
	icon.siz->height_inc=icon.siz->width_inc=0;
	XSetIconSizes(dp,win,icon.siz,1);
	XFree(icon.siz);
	if (!(icon.win=XCreateSimpleWindow(dp,rootw,0,0,
			icon.attr.width,icon.attr.height,0,bdpen,bgpen))) iofail();
	XpmFreeAttributes(&icon.attr);
	XSetWindowBackgroundPixmap(dp,icon.win,icon.data);
	XFreePixmap(dp,icon.data);
	if (icon.mask) {
		XShapeCombineMask(dp,icon.win,ShapeBounding,0,0,icon.mask,ShapeSet);
		XFreePixmap(dp,icon.mask);
		}
	XClearWindow(dp,icon.win);
	if (!(wmh=XAllocWMHints())) iofail();
	wmh->flags=StateHint|IconWindowHint;
	wmh->initial_state=res_bool(res_string(10,".IconStartup"))?IconicState:NormalState;
	wmh->icon_window=icon.win;
	if (!(clsh=XAllocClassHint())) iofail();
	clsh->res_name=pname;
	clsh->res_class=(char *)class;
	if (!(s=res_string(16,".Title"))) s=(char *)class;
	XmbSetWMProperties(dp,win,s,class,argv,*argcp,sizh,wmh,clsh);
	XFree(sizh); XFree(wmh); XFree(clsh);
	wina.backing_store=WhenMapped;
	wina.event_mask=ButtonPressMask|KeyPressMask|StructureNotifyMask|ExposureMask;
	wina.cursor=curs=XCreateFontCursor(dp,XC_watch);
	XChangeWindowAttributes(dp,win,CWBackingStore|CWEventMask|CWCursor,&wina);
	atdel=XInternAtom(dp,"WM_DELETE_WINDOW",False);
	XSetWMProtocols(dp,win,&atdel,1);
	XMapWindow(dp,win);
	XFlush(dp);
	gc=DefaultGCOfScreen(scr);
}
